import asyncio
import functools

import spotipy
from spotipy.oauth2 import CacheFileHandler

from .. import fhem, utils
from .. import generic


class spotify(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        # Spotipy Spotify instance
        self.spotipy = None
        # Spotipy PKCE authenticator instance
        self.spotipy_pkce = None
        # Available Spotify Connect devices
        self.available_devices = {}
        self._last_data = {}
        self.spotipy_scope = (
            ""
            # Spotify scopes: https://developer.spotify.com/documentation/general/guides/scopes/
            # Listening history
            + "user-read-recently-played "
            + "user-top-read "
            + "user-read-playback-position "
            # Spotify connect
            + "user-read-playback-state "
            + "user-modify-playback-state "
            + "user-read-currently-playing "
            # Playback
            + "streaming "
            # Playlist
            + "playlist-read-private "
            + "playlist-read-collaborative "
            # Follow
            + "user-follow-read "
            # Library
            + "user-library-read "
            # User
            + "user-read-email "
            + "user-read-private"
        )
        self.loggedin = False
        attr_config = {
            "update_status_interval": {
                "default": 60,
                "format": "int",
                "help": "Change interval, default is 60s.",
            },
            "update_devices_interval": {
                "default": 300,
                "format": "int",
                "help": "Change interval, default is 300s.",
            },
            "default_device": {"default": None, "help": "Set default device ID."},
            "spotify_connect": {
                "default": "off",
                "options": "on,off",
                "help": "Activate FHEM Connect Player within FHEM Web. Just open Spotify app and stream to FHEM Web Player.",
            },
        }
        self.set_attr_config(attr_config)

        self.set_config = {
            "play": {
                "args": ["uri", "deviceid"],
                "function": "set_command",
                "params": {
                    "uri": {"default": None, "optional": False},
                    "deviceid": {"default": None, "optional": True},
                },
            },
            "pause": {
                "args": ["deviceid"],
                "function": "set_command",
                "params": {"deviceid": {"default": None, "optional": True}},
            },
            "next_track": {
                "args": ["deviceid"],
                "function": "set_command",
                "params": {"deviceid": {"default": None, "optional": True}},
            },
            "previous_track": {
                "args": ["deviceid"],
                "function": "set_command",
                "params": {"deviceid": {"default": None, "optional": True}},
            },
            "authcode": {"args": ["code"]},
            "update_devices": {},
            "status": {"function": "set_command"},
            "shuffle": {
                "args": ["onoff"],
                "params": {
                    "onoff": {"default": True, "optional": True},
                },
                "options": "on,off",
                "function": "set_command",
            },
            "volume": {
                "args": ["vol"],
                "params": {
                    "vol": {"optional": False},
                },
                "options": "slider,0,1,100",
                "function": "set_command",
            },
            "transfer_playback": {
                "args": ["deviceid"],
                "function": "set_command",
            },
        }
        self.set_set_config(self.set_config)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "disconnected")
        await fhem.readingsEndUpdate(hash, 1)
        await self.set_auth_url()
        self.create_async_task(self.connect_spotipy())

    async def FW_detailFn(self, hash, args, argsh):
        spotify_script = ""
        if self._attr_spotify_connect == "on":
            spotify_script = """
                <script src="https://sdk.scdn.co/spotify-player.js"></script>
                <script>
                    window.onSpotifyWebPlaybackSDKReady = () => {
                    const token = '###SPOTIFY_ACCESS_TOKEN###';
                    const player = new Spotify.Player({
                        name: 'FHEM Web Player',
                        getOAuthToken: cb => { cb(token); }
                    });

                    // Error handling
                    player.addListener('initialization_error', ({ message }) => { console.error(message); });
                    player.addListener('authentication_error', ({ message }) => { console.error(message); });
                    player.addListener('account_error', ({ message }) => { console.error(message); });
                    player.addListener('playback_error', ({ message }) => { console.error(message); });

                    // Playback status updates
                    player.addListener('player_state_changed', state => { 
                        console.log(state);
                        var div = document.getElementsByClassName("makeSelect")[0];
                        var dev = div.getAttribute("dev");
                        FW_cmd(FW_root + '?cmd=set ' + dev + ' status&XHR=1');
                    });

                    // Ready
                    player.addListener('ready', ({ device_id }) => {
                        console.log('Ready with Device ID', device_id);
                        $('<div id="idSpotifyConnect" class="makeTable help"></div>')
                            .insertBefore("div.makeSelect");
                        $("div#idSpotifyConnect").html("Spotify Connect active with device id: " + device_id + "<br>");
                    });

                    // Not Ready
                    player.addListener('not_ready', ({ device_id }) => {
                        console.log('Device ID has gone offline', device_id);
                    });

                    // Connect to the player!
                    player.connect();
                    };
                </script>
            """
            access_token = self.spotipy_pkce.get_cached_token()
            if access_token is not None:
                spotify_script = spotify_script.replace(
                    "###SPOTIFY_ACCESS_TOKEN###", access_token["access_token"]
                )
        detailfn = await super().FW_detailFn(hash, args, argsh)
        return spotify_script + str(detailfn)

    async def set_auth_url(self):
        # not sure if this is useful...
        x = "e92855a009"
        y = "e74eb69ba6609d3bfd7d"
        z = 90 + 6
        handler = CacheFileHandler(cache_path=f".{self.hash['NAME']}_spotify_token")
        self.spotipy_pkce = spotipy.oauth2.SpotifyPKCE(
            f"{x}{y}{str(z)}",
            "https://europe-west1-fhem-ga-connector.cloudfunctions.net/codelanding/start",
            scope=self.spotipy_scope,
            cache_handler=handler,
        )
        url = self.spotipy_pkce.get_authorize_url()
        url = (
            '<html><a href="'
            + url
            + '" target="_blank">Connect Spotify account (new window/tab)</a><br></html>'
        )
        await fhem.readingsSingleUpdate(self.hash, "login", url, 1)

    async def set_authcode(self, hash, params):
        code = params["code"]
        self.create_async_task(self.handle_authcode(code))

    async def handle_authcode(self, code):
        access_token = await utils.run_blocking(
            functools.partial(
                self.spotipy_pkce.get_access_token, code=code, check_cache=False
            )
        )
        await self.connect_spotipy()

    async def connect_spotipy(self):
        if self.spotipy_pkce.get_cached_token() is None:
            await fhem.readingsSingleUpdate(self.hash, "state", "login required", 1)
            return

        if self.spotipy is None:
            self.spotipy = spotipy.Spotify(auth_manager=self.spotipy_pkce)
            user = await utils.run_blocking(
                functools.partial(self.spotipy.current_user)
            )
            if user is not None:
                self.loggedin = True
                await fhem.readingsSingleUpdate(
                    self.hash,
                    "user",
                    user["display_name"] + " (" + user["email"] + ")",
                    1,
                )
                await fhem.readingsSingleUpdate(self.hash, "state", "connected", 1)
                await self.update_devices()
                self.create_async_task(self.update_playback_loop())
                self.create_async_task(self.update_devices_loop())

    async def update_playback_loop(self):
        while True:
            try:
                await self.update_playback()
            except Exception:
                self.logger.exception("Failed to update playback")
            await asyncio.sleep(self._attr_update_status_interval)

    async def update_devices_loop(self):
        while True:
            try:
                await self.update_devices()
            except Exception:
                self.logger.exception("Failed to update devices")
            await asyncio.sleep(self._attr_update_devices_interval)

    async def set_update_devices(self, hash, params):
        self.create_async_task(self.update_devices())

    async def update_devices(self):
        if not self.loggedin:
            return
        # Query spotify for active devices
        devices_available = await utils.run_blocking(
            functools.partial(self.spotipy.devices)
        )

        self.available_devices = {}
        await fhem.CommandDeleteReading(self.hash, f"{self.hash['NAME']} device_.*")
        # update device list
        await fhem.readingsBeginUpdate(self.hash)
        for device in devices_available["devices"]:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, "device_" + device["id"], device["name"]
            )
            self.available_devices[device["id"]] = device
            if self._attr_default_device is None:
                self._attr_default_device = device["id"]
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash,
                    "default_device",
                    device["name"] + " (" + device["id"] + ")",
                )
        await fhem.readingsEndUpdate(self.hash, 1)

    async def set_attr_default_device(self, hash):
        if self._attr_default_device in self.available_devices:
            await fhem.readingsSingleUpdateIfChanged(
                self.hash,
                "default_device",
                self._attr_default_device
                + " ("
                + self.available_devices[self._attr_default_device]["name"]
                + ")",
                1,
            )

    async def set_command(self, hash, params):
        if not self.loggedin:
            return "Please login first"

        if params["cmd"] == "pause":
            utils.run_blocking_task(
                functools.partial(self.spotipy.pause_playback, params["deviceid"])
            )
        elif params["cmd"] == "play":
            if "track" in params["uri"]:
                utils.run_blocking_task(
                    functools.partial(
                        self.spotipy.start_playback,
                        device_id=params["deviceid"],
                        uris=[params["uri"]],
                    )
                )
            else:
                utils.run_blocking_task(
                    functools.partial(
                        self.spotipy.start_playback,
                        device_id=params["deviceid"],
                        context_uri=params["uri"],
                    )
                )
        elif params["cmd"] == "next_track":
            utils.run_blocking_task(
                functools.partial(self.spotipy.next_track, params["deviceid"])
            )
        elif params["cmd"] == "previous_track":
            utils.run_blocking_task(
                functools.partial(self.spotipy.previous_track, params["deviceid"])
            )
        elif params["cmd"] == "status":
            self.create_async_task(self.update_playback())
        elif params["cmd"] == "shuffle":
            utils.run_blocking_task(
                functools.partial(self.spotipy.shuffle, params["onoff"] == "on")
            )
        elif params["cmd"] == "volume":
            utils.run_blocking_task(
                functools.partial(self.spotipy.volume, params["vol"])
            )
        elif params["cmd"] == "transfer_playback":
            utils.run_blocking_task(
                functools.partial(self.spotipy.transfer_playback, params["deviceid"])
            )

    async def update_playback(self):
        try:
            status = await utils.run_blocking(
                functools.partial(self.spotipy.current_playback)
            )
            flat_status = utils.flatten_json(status)
            if self._last_data:
                del_readings = set(self._last_data) - set(flat_status)
            else:
                del_readings = {}
                await fhem.CommandDeleteReading(
                    self.hash, self.hash["NAME"] + " current_.*"
                )

            await fhem.readingsBeginUpdate(self.hash)
            for status_name in flat_status:
                reading = "current_" + status_name
                if "available_markets" in status_name:
                    continue
                if "device_volume_percent" == status_name:
                    reading = "volume"
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, reading, flat_status[status_name]
                )
            await fhem.readingsEndUpdate(self.hash, 1)

            # delete old readings which were not updated
            for del_reading in del_readings:
                await fhem.CommandDeleteReading(
                    self.hash, self.hash["NAME"] + " current_" + del_reading
                )
            self._last_current = flat_status
        except Exception:
            self.logger.exception("Failed to update status")



import asyncio
import functools
import logging

from .. import fhem
from .. import utils as fpyutils
from .. import generic
from .nespresso import NespressoDetect


class nespresso_ble(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.nespressodetect = None
        self.task = None
        self.auth = None
        logging.getLogger("pygatt.backends.gatttool.gatttool").setLevel(logging.ERROR)
        # logging.getLogger("nespresso_ble").setLevel(logging.DEBUG)
        self.set_conf_list = {
            "authkey": {"args": ["authkey"]},
            "brew": {
                "args": ["coffee_type", "temperature"],
                "params": {
                    "temperature": {"default": "high", "optional": True},
                    "coffee_type": {"default": "lungo", "optional": True},
                },
                "help": "1. parameter: ristretto, espresso, lungo, hotwater, americano<br>2. parameter: Temperature with values low, mid, high",
            },
            "easybrew": {
                "args": ["coffee_type"],
                "options": "ristretto,espresso,lungo,hotwater,americano",
            },
            "recipe": {"help": "Not yet supported"},
            "updateStatus": {},
        }
        self.set_set_config(self.set_conf_list)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 4:
            return "Usage: define devicename fhempy nespresso_ble <MAC> [<AUTHKEY>]"
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "offline")
        await fhem.readingsEndUpdate(hash, 1)
        self.mac = args[3]
        hash["MAC"] = args[3]

        # check if there is already an authkey
        if len(args) > 4:
            self.auth = args[4]
        else:
            self.auth = await fhem.ReadingsVal(self.hash["NAME"], "authkey", "")

        if self.auth != "":
            self.auth = args[4]
            self.nespressodetect = NespressoDetect(self.auth, self.mac)
            self.nespressodetect.set_keep_connected(True)
            self.task = self.create_async_task(self.update_status_task())
        return ""

    async def update_status_task(self):
        while True:
            if self.nespressodetect:
                await self.update_status()
            await asyncio.sleep(300)

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        if self.auth and "authkey" in self.set_conf_list:
            del self.set_conf_list["authkey"]
            self.set_set_config(self.set_conf_list)
        return await super().Set(hash, args, argsh)

    async def set_authkey(self, hash, params):
        self.auth = params["authkey"]
        self.set_set_config(self.set_conf_list)
        await fhem.readingsSingleUpdateIfChanged(self.hash, "authkey", self.auth, 1)
        if self.task:
            self.task.cancel()
        self.nespressodetect = NespressoDetect(self.auth, self.mac)
        self.nespressodetect.set_keep_connected(True)
        self.task = self.create_async_task(self.update_status_task())

    async def set_easybrew(self, hash, params):
        params["temperature"] = "medium"
        await self.set_brew(hash, params)

    async def set_brew(self, hash, params):
        try:
            coffee_type = params["coffee_type"]
            temp = params["temperature"]
            fpyutils.run_blocking_task(
                functools.partial(
                    self.nespressodetect.make_coffee, self.mac, temp, coffee_type
                )
            )
        except Exception:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)

    async def set_updateStatus(self, hash, params):
        self.create_async_task(self.update_status())

    async def update_status(self):
        await fpyutils.run_blocking(functools.partial(self.blocking_update_status))

        if self.device_info:
            for mac, dev in self.device_info.items():
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "manufacturer", dev.manufacturer, 1
                )
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "serial_nr", dev.serial_nr, 1
                )
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "model_nr", dev.model_nr, 1
                )
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "device_name", dev.device_name, 1
                )
                if dev.device_name:
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "online", 1
                    )
                else:
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "offline", 1
                    )
        else:
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)

        if self.sensors_data:
            for mac, data in self.sensors_data.items():
                for name, val in data.items():
                    await fhem.readingsSingleUpdateIfChanged(self.hash, name, val, 1)

    def blocking_update_status(self):
        self.logger.debug("nespresso_ble updatestatus")
        try:
            self.device_info = self.nespressodetect.get_info()
            self.nespressodetect.get_sensors()
            self.sensors_data = self.nespressodetect.get_sensor_data()

        except Exception:
            self.logger.exception("Failed to update status")
            self.sensors_data = None
            self.device_info = None



import asyncio
import pytest
import logging
import os
from tests.utils import mock_fhem
from fhempy.lib.xiaomi_gateway3.xiaomi_gateway3 import xiaomi_gateway3
from fhempy.lib.xiaomi_gateway3_device.xiaomi_gateway3_device import (
    xiaomi_gateway3_device,
)


@pytest.mark.asyncio
async def test_general(mocker):
    """
    This test requires following environment variables defined
    in pytest.ini:
        XIAOMI_GATEWAY3_IP
        XIAOMI_GATEWAY3_TOKEN
        XIAOMI_GATEWAY3_MAC
        XIAOMI_GATEWAY3_lumiweatherv1=did
        XIAOMI_GATEWAY3_sensormagnetaq2=did
        XIAOMI_GATEWAY3_sensormotionv2=did
    """
    # prepare
    mock_fhem.mock_module(mocker)
    mocker.patch("socket.socket.connect_ex", return_value=0)

    testhash = {"NAME": "testdevice", "FHEMPYTYPE": "testtype"}
    logger = logging.getLogger(__name__)
    xg_module = xiaomi_gateway3(logger)
    await xg_module.Define(
        testhash,
        [
            "testdevice",
            "fhempy",
            "xiaomi_gateway3",
            os.environ["XIAOMI_GATEWAY3_IP"],
            os.environ["XIAOMI_GATEWAY3_TOKEN"],
        ],
        {},
    )

    mocker.patch(
        "fhempy.lib.fhem_pythonbinding.getFhemPyDeviceByName", return_value=xg_module
    )
    # wait for connection setup and device creation
    await asyncio.sleep(3)
    # remove duplicates
    mock_fhem.command_define = list(dict.fromkeys(mock_fhem.command_define))
    assert (
        mock_fhem.command_define.index(
            f"lumigatewaymgl03_0x{os.environ['XIAOMI_GATEWAY3_MAC']} "
            f"fhempy xiaomi_gateway3_device testdevice "
            f"lumi.0x{os.environ['XIAOMI_GATEWAY3_MAC']}"
        )
        > -1
    )

    for dev_define in mock_fhem.command_define:
        defarr = dev_define.split(" ")
        devname = defarr[0]
        devhash = {"NAME": devname}
        dev_module = xiaomi_gateway3_device(logger)
        await dev_module.Define(devhash, defarr, {})

    await asyncio.sleep(5)

    # weather test
    weather_readings = mock_fhem.readings[
        "lumiweatherv1_0x" + os.environ["XIAOMI_GATEWAY3_lumiweatherv1"]
    ]
    assert weather_readings["state"] == "online"
    assert (
        weather_readings["did"] == "lumi." + os.environ["XIAOMI_GATEWAY3_lumiweatherv1"]
    )
    assert weather_readings["mac"] == "0x" + os.environ["XIAOMI_GATEWAY3_lumiweatherv1"]
    assert weather_readings["model"] == "lumi.weather.v1"
    assert weather_readings["type"] == "zigbee"
    # assert weather_readings["zb_ver"] == "1.2"
    assert float(weather_readings["temperature"]) > 0
    assert float(weather_readings["humidity"]) > 0
    assert float(weather_readings["pressure"]) > 0
    assert int(weather_readings["battery"]) > 0
    assert weather_readings["online"] is True
    assert weather_readings["device_manufacturer"] == "Aqara"
    assert weather_readings["device_model"] == "lumi.weather WSDCGQ11LM"

    # sensor magnet test
    sensor_readings = mock_fhem.readings[
        "lumisensormagnetaq2_0x" + os.environ["XIAOMI_GATEWAY3_sensormagnetaq2"]
    ]
    assert sensor_readings["state"] == "online"
    assert (
        sensor_readings["did"]
        == "lumi." + os.environ["XIAOMI_GATEWAY3_sensormagnetaq2"]
    )
    assert (
        sensor_readings["mac"] == "0x" + os.environ["XIAOMI_GATEWAY3_sensormagnetaq2"]
    )
    assert sensor_readings["model"] == "lumi.sensor_magnet.aq2"
    assert sensor_readings["type"] == "zigbee"
    # assert sensor_readings["zb_ver"] == "1.2"
    assert int(sensor_readings["contact"]) >= 0
    assert int(sensor_readings["battery"]) > 0
    assert sensor_readings["online"] is True
    assert sensor_readings["device_manufacturer"] == "Aqara"
    assert sensor_readings["device_model"] == "lumi.sensor_magnet.aq2 MCCGQ11LM"

    # sensor motion test
    sensor_readings = mock_fhem.readings[
        "lumisensormotionv2_0x" + os.environ["XIAOMI_GATEWAY3_sensormotionv2"]
    ]
    assert sensor_readings["state"] == "online"
    assert (
        sensor_readings["did"] == "lumi." + os.environ["XIAOMI_GATEWAY3_sensormotionv2"]
    )
    assert sensor_readings["mac"] == "0x" + os.environ["XIAOMI_GATEWAY3_sensormotionv2"]
    assert sensor_readings["model"] == "lumi.sensor_motion.v2"
    assert sensor_readings["type"] == "zigbee"
    # assert sensor_readings["zb_ver"] == "1.2"
    assert int(sensor_readings["battery"]) > 0
    assert sensor_readings["online"] is True
    assert sensor_readings["device_manufacturer"] == "Xiaomi"
    assert sensor_readings["device_model"] == "lumi.sensor_motion RTCGQ01LM"



import ast
import asyncio
import functools
import json
import re

from tinytuya import Cloud, deviceScan

from .. import fhem, generic, utils
from . import mappings, pytuya


class tuya(generic.FhemModule, pytuya.TuyaListener):
    def __init__(self, logger):
        super().__init__(logger)
        self._connected_device = None
        self.tuya_cloud = None
        self.tt_key = ""
        self.tt_secret = ""
        self.last_status = None
        self.create_device_list = []
        self.update_lock = asyncio.Lock()

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 7:
            return (
                "Usage: define wifi_plug fhempy tuya"
                " setup <API_KEY> <API_SECRET> <DEVICE_ID> [<REGION>=eu]<br>"
                "OR if you want to define only one device with existing local key<br>"
                " <PRODUCTID> <DEVICE_ID> <IP> <LOCAL_KEY> "
                "[<VERSION>=3.3] [<API_KEY>] [<API_SECRET>]"
            )

        self.tt_type = args[3]
        if self.tt_type == "setup":
            self.tt_key = args[4]
            self.tt_secret = args[5]
            self.tt_did = args[6]
            if len(args) == 8:
                self.tt_region = args[7]
            else:
                self.tt_region = "eu"
            self.set_set_config({"scan_devices": {}})
            await fhem.readingsSingleUpdate(self.hash, "state", "ready", 1)
            # set internal
            hash["DEVICEID"] = 0
            hash["API_KEY"] = self.tt_key
            hash["API_SECRET"] = self.tt_secret
            hash["REGION"] = self.tt_region
            self.create_async_task(self.setup_cloud())
            return

        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)
        self.tt_region = "eu"
        self.tt_did = args[4]
        self.tt_ip = args[5]
        self.tt_localkey = args[6]
        if len(args) >= 8:
            self.tt_version = float(args[7])
        else:
            self.tt_version = 3.3
        if len(args) == 10:
            self.tt_key = args[8]
            self.tt_secret = args[9]

        if self.tt_ip == "offline":
            await fhem.readingsSingleUpdate(
                self.hash, "state", "Change DEF and use IP instead of 'offline'", 1
            )
            return

        # set internal
        hash["DEVICEID"] = self.tt_did
        # set attributes
        self.attr_config = {
            "tuya_spec_functions": {"default": ""},
            "tuya_spec_status": {"default": ""},
        }
        self.set_attr_config(self.attr_config)
        # this is needed to set default values
        await utils.handle_define_attr(self.attr_config, self, hash)

        await self.setup_cloud()

        # create device
        self.create_async_task(self.create_device())

    async def setup_cloud(self):
        # create cloud device for cloud calls
        if self.tt_key and self.tt_secret:
            self.tuya_cloud = await utils.run_blocking(
                functools.partial(
                    Cloud,
                    self.tt_region,
                    self.tt_key,
                    self.tt_secret,
                    self.tt_did,
                )
            )

    async def _create_mapping_dev(self):
        self.tuya_spec_functions = []
        self.tuya_spec_status = []
        schema = mappings.knownSchemas[self.tt_type]["schema"]
        for schema_part in schema:
            # rename to values
            schema_part["values"] = schema_part["property"]
            schema_part["type"] = schema_part["values"]["type"]
            schema_part["desc"] = ""
            schema_part["values"] = json.dumps(schema_part["values"])
            if schema_part["type"] == "bool":
                schema_part["type"] = "Boolean"
            elif schema_part["type"] == "value":
                schema_part["type"] = "Integer"
            elif schema_part["type"] == "enum":
                schema_part["type"] = "Enum"
            elif schema_part["type"] == "string":
                schema_part["type"] = "String"
            else:
                continue
            self.tuya_spec_status.append(schema_part)
            if schema_part["mode"] == "rw":
                self.tuya_spec_functions.append(schema_part)

        await self._generate_set()

        await self.setup_connection()
        status = await self._connected_device.status()
        await self.update_readings(status)

    async def set_attr_dp(self, hash):
        # check defined dp_Xs and their value
        # set set_conf based on the set dp_Xs
        dp_conf = {}
        for attr_name in self.__dict__:
            if attr_name.find("_attr_dp_") != -1:
                dp_id = re.search(r"\d+$", attr_name)[0]
                dp_conf[dp_id] = self.__dict__[attr_name]
                for fct in self.tuya_spec_functions:
                    if fct["code"] == dp_conf[dp_id]:
                        fct["id"] = int(dp_id)
                for status in self.tuya_spec_status:
                    if status["code"] == dp_conf[dp_id]:
                        status["id"] = int(dp_id)

        await self._generate_set()

    async def _generate_set(self):
        set_conf = {}
        for fct in self.tuya_spec_functions:
            if "id" not in fct:
                # no id configured for this function
                continue
            if fct["type"] == "Boolean":
                if fct["code"] == "switch_1":
                    set_conf["on"] = {
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
                    set_conf["off"] = {
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
                else:
                    set_conf[fct["code"]] = {
                        "options": "on,off",
                        "args": ["onoff"],
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
            elif fct["type"] == "Enum":
                options = json.loads(fct["values"])["range"]
                set_conf[fct["code"]] = {
                    "options": ",".join(options),
                    "args": ["new_val"],
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_other_types",
                }
            elif fct["type"] == "Integer":
                spec = json.loads(fct["values"])
                slider = f"slider,{spec['min']},{spec['step']},{spec['max']}"
                set_conf[fct["code"]] = {
                    "options": slider,
                    "args": ["new_val"],
                    "params": {"new_val": {"format": "int"}},
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_other_types",
                }
            elif fct["type"] == "String":
                set_conf[fct["code"]] = {
                    "args": ["new_val"],
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_other_types",
                }
            elif fct["type"] == "Json":
                pass

        self.set_set_config(set_conf)

    async def set_create_device(self, hash, params):
        nameid = params["name"]
        for dev in self.create_device_list:
            if nameid == dev["name_esc"] + "_" + dev["device_id"]:
                await fhem.CommandDefine(
                    self.hash,
                    (
                        f"{dev['name_esc']}_{dev['device_id']} fhempy tuya "
                        f"{dev['productid']} {dev['device_id']} {dev['ip']} "
                        f"{dev['local_key']} {dev['version']} "
                        f"{self.tt_key} {self.tt_secret}"
                    ),
                )

    async def set_boolean(self, hash, params):
        switch_id = params["function_param"]["id"]
        onoff = False
        if "onoff" in params:
            if params["onoff"] == "on":
                onoff = True
        else:
            if params["cmd"] == "on":
                onoff = True
        if self._connected_device:
            await self._connected_device.set_dp(onoff, switch_id)

    async def set_other_types(self, hash, params):
        index = params["function_param"]["id"]
        new_val = params["new_val"]
        if self._connected_device:
            await self._connected_device.set_dp(new_val, index)

    # check if tuya_spec_functions and tuya_spec_status attr is set?
    async def check_tuya_attributes(self):
        self.tuya_spec_functions = await fhem.AttrVal(
            self.hash["NAME"], "tuya_spec_functions", "[]"
        )
        self.tuya_spec_status = await fhem.AttrVal(
            self.hash["NAME"], "tuya_spec_status", "[]"
        )
        self.tuya_spec_functions = ast.literal_eval(self.tuya_spec_functions)
        self.tuya_spec_status = ast.literal_eval(self.tuya_spec_status)

    # get functions/status from tuya
    async def get_tuya_dev_specification(self):
        # Get specification
        resp = await utils.run_blocking(
            functools.partial(self.tuya_cloud.getdps, self.tt_did)
        )
        return resp["result"]

    async def get_tuya_dev_info(self):
        resp = await utils.run_blocking(
            functools.partial(self.tuya_cloud.getdevices, True)
        )
        for dev in resp["result"]:
            if dev["id"] == self.tt_did:
                return dev
        return {}

    async def get_tuya_dev_description(self):
        # Get function description
        response_dict = await utils.run_blocking(
            functools.partial(self.tuya_cloud.getfunctions, self.tt_did)
        )
        fct_desc = response_dict["result"]["functions"]
        fct_code_desc = {}
        for fct in fct_desc:
            fct_code_desc[fct["code"]] = fct["desc"]
        return fct_code_desc

    async def async_status_updated(self, status):
        await self.update_readings(status)

    def status_updated(self, status):
        self.create_async_task(self.async_status_updated(status))

    def disconnected(self):
        self.create_async_task(self.async_disconnected())

    async def _add_desc_to_spec(self, spec_fcts, desc):
        for spec in spec_fcts:
            if spec["code"] in desc:
                spec["desc"] = desc[spec["code"]]
        return spec_fcts

    async def retrieve_tuya_specs(self):
        # retrieve cloud codes
        spec = await self.get_tuya_dev_specification()
        self.tuya_spec_functions = spec["functions"]
        self.tuya_spec_status = spec["status"]
        desc = await self.get_tuya_dev_description()
        self.tuya_spec_functions = await self._add_desc_to_spec(
            self.tuya_spec_functions, desc
        )
        await fhem.CommandAttr(
            self.hash,
            (
                f"{self.hash['NAME']} tuya_spec_functions "
                f"{str(self.tuya_spec_functions)}"
            ),
        )
        await fhem.CommandAttr(
            self.hash,
            f"{self.hash['NAME']} tuya_spec_status {str(self.tuya_spec_status)}",
        )

    async def _create_cloudmapping_dev(self):
        await fhem.readingsSingleUpdate(self.hash, "state", "Initializing...", 1)
        # check if attributes are set
        await self.check_tuya_attributes()
        if len(self.tuya_spec_functions) == 0 and len(self.tuya_spec_status) == 0:
            await self.retrieve_tuya_specs()

        # create attributes dp_1...X
        # add options to attributes to select cloud codes
        # e.g.
        # dp_1 = "switch_1"
        # dp_42 = "switch_overcharge"
        # connect to device
        info = await self.get_tuya_dev_info()
        await self.update_info_readings(info)

        await self.setup_connection()

        status = await self._connected_device.status()

        await self.prepare_attributes(status)
        await self.update_readings(status)

    async def prepare_attributes(self, status):
        options = []
        for opt in self.tuya_spec_status:
            options.append(opt["code"])

        dps = []
        for dp in list(status):
            dps.append(f"dp_{int(dp):02d}")

        attr_conf = {}
        for dp in dps:
            attr_conf[dp] = {
                "function": "set_attr_dp",
                "default": "",
                "options": ",".join(options),
            }
        self.attr_config.update(attr_conf)
        self.set_attr_config(self.attr_config)
        await utils.handle_define_attr(self.attr_config, self, self.hash)

        # if spec contains dp_id
        for spec in self.tuya_spec_status:
            if "dp_id" in spec:
                dp = spec["dp_id"]
                code = spec["code"]
                val = await fhem.AttrVal(self.hash["NAME"], f"dp_{int(dp):02d}", "")
                if val == "":
                    await fhem.CommandAttr(
                        self.hash,
                        f"{self.hash['NAME']} dp_{int(dp):02d} {code}",
                    )
        for spec in self.tuya_spec_functions:
            if "dp_id" in spec:
                dp = spec["dp_id"]
                code = spec["code"]
                val = await fhem.AttrVal(self.hash["NAME"], f"dp_{int(dp):02d}", "")
                if val == "":
                    await fhem.CommandAttr(
                        self.hash,
                        f"{self.hash['NAME']} dp_{int(dp):02d} {code}",
                    )
        await self.set_attr_dp(self.hash)

    async def setup_connection(self):
        while True:
            try:
                self._connected_device = await pytuya.connect(
                    self.tt_ip,
                    self.tt_did,
                    self.tt_localkey,
                    self.tt_version,
                    self,
                    timeout=15,
                )
                break
            except Exception:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "offline", 1
                )
                self.logger.exception("Failed to connect to device")
                await asyncio.sleep(1)

    async def create_device(self):
        try:
            if self.tt_key != "" and self.tt_secret != "":
                await self._create_cloudmapping_dev()
            elif self.tt_type in mappings.knownSchemas:
                await self._create_mapping_dev()
            else:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash,
                    "state",
                    "Please use API_KEY and API_SECRET",
                )
        except Exception:
            self.logger.exception("Failed create_device")

    async def async_disconnected(self):
        await fhem.readingsSingleUpdate(self.hash, "state", "offline", 1)
        self._connected_device = None
        await asyncio.sleep(5)
        await self.setup_connection()

    def convert(self, value, schema):
        if schema["type"] == "Integer":
            values = json.loads(schema["values"])
            return value / (10 ** values["scale"])
        elif schema["type"] == "Boolean":
            if value is True:
                return "on"
            return "off"
        return value

    async def update_info_readings(self, info_dict):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            for reading in info_dict:
                if reading == "status":
                    continue
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, reading, info_dict[reading]
                )
        except Exception:
            self.logger.exception("Failed to update info readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_readings(self, status):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            stateused = False
            for dp in status:
                found = False
                for st in self.tuya_spec_status:
                    if "id" in st and st["id"] == int(dp):
                        found = True
                        reading = st["code"]
                        if st["code"] == "switch_1":
                            reading = "state"
                            stateused = True
                        await fhem.readingsBulkUpdateIfChanged(
                            self.hash,
                            reading,
                            self.convert(status[dp], st),
                        )
                        break

                if not found:
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash,
                        f"dp_{int(dp):02d}",
                        status[dp],
                    )
            if not stateused:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "online")
        except Exception:
            self.logger.exception("Failed to update readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    async def Undefine(self, hash):
        if self._connected_device:
            await self._connected_device.close()
            self._connected_device = None
        await super().Undefine(hash)

    # The following code is only for setup/scan process via tuya cloud
    async def set_scan_devices(self, hash, params):
        self.create_async_task(self._scan_devices())

    async def _scan_devices(self):
        json_data = await utils.run_blocking(
            functools.partial(self.tuya_cloud.getdevices, verbose=True)
        )

        # Filter to only Name, ID and Key, product_id, icon
        tuyadevices = []
        for i in json_data["result"]:
            item = {}
            item["name"] = i["name"].strip()
            item["id"] = i["id"]
            item["ip"] = i["ip"]
            item["sub"] = i["sub"]
            item["key"] = i["local_key"]
            item["product_id"] = i["product_id"]
            item["icon"] = i["icon"]
            await fhem.readingsBeginUpdate(self.hash)
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_name", item["name"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_id", item["id"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_localkey", item["key"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_productid", item["product_id"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                item["id"] + "_icon",
                "https://images.tuyaeu.com/" + item["icon"],
            )
            await fhem.readingsEndUpdate(self.hash, 1)
            tuyadevices.append(item)

        # Display device list
        self.logger.debug("Device Listing")
        output = json.dumps(tuyadevices, indent=4)  # sort_keys=True)
        self.logger.debug(output)
        await fhem.readingsSingleUpdate(
            self.hash,
            "state",
            f"found {len(tuyadevices)} devices, start local scan...",
            1,
        )

        # scan local devices to get IP
        self.logger.debug("Scan local devices...")
        devices = await utils.run_blocking(functools.partial(deviceScan, False, 20))

        def getIP(d, gwid):
            for ip in d:
                if gwid == d[ip]["gwId"]:
                    return (ip, d[ip]["version"])
            return (0, 0)

        self.logger.debug("Polling local devices...")
        count_found = 0
        self.create_device_list = []
        for i in tuyadevices:
            name = i["name"]
            id = i["id"]
            (ip, ver) = getIP(devices, i["id"])
            local_key = i["key"]
            productid = i["product_id"]

            # do not create device if no IP was discovered
            if ver == 0 and i["sub"] is True:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, f"{id}_ip", "offline", 1
                )
                continue

            if ver == 0:
                # device not discovered
                ip = "offline"
                ver = "3.3"
            else:
                count_found += 1

            await fhem.readingsSingleUpdateIfChanged(self.hash, f"{id}_ip", ip, 1)
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, f"{id}_version", str(ver), 1
            )

            # create FHEM device
            self.create_device_list.append(
                {
                    "name": name,
                    "name_esc": name.replace(" ", "_"),
                    "device_id": id,
                    "productid": productid,
                    "ip": ip,
                    "local_key": local_key,
                    "version": ver,
                }
            )

        # devices which can be created from setup
        set_conf = {"scan_devices": {}}
        set_conf["create_device"] = {"args": ["name"], "options": ""}
        options = []
        for dev in self.create_device_list:
            options.append(dev["name_esc"] + "_" + dev["device_id"])
        set_conf["create_device"]["options"] = ",".join(options)
        self.set_set_config(set_conf)

        await fhem.readingsSingleUpdate(
            self.hash,
            "state",
            f"{count_found} devices found localy",
            1,
        )
